import os
import uuid
from xml.dom import minidom
from copy import deepcopy

import pyvista
import tetgen
import svv
import tqdm
import numpy
import pymeshfix
import vtk

import svv.tree.tree
from svv.simulation.utils.extract_faces import extract_faces
from svv.domain.routines.boolean import boolean
from svv.utils.remeshing.remesh import remesh_surface, remesh_volume
from svv.simulation.general_parameters import GeneralSimulationParameters
from svv.simulation.mesh import GeneralMesh
from svv.simulation.fluid.fluid_equation import FluidEquation
from svv.simulation.utils.boundary_layer import BoundaryLayer


from svv.simulation.fluid.rom import one_d
from svv.simulation.fluid.rom import zero_d
from svv.simulation.fluid.rom.zero_d.zerod_forest import export_0d_simulation
from svv.simulation.fluid.rom.zero_d import project_solution


class Simulation(object):
    def __init__(self, synthetic_object, name=None, directory=None):
        """
        The Simulation class defines a simulation object that
        is used to generate the files to run a physics simulation
        using a synthetic vascular network.
        """
        self.synthetic_object = synthetic_object
        if name is None:
            name = "simulations_" + uuid.uuid4().hex
        if directory is None:
            directory = os.getcwd()
        self.file_path = os.path.join(directory, name)
        self.tissue_domain_faces = []
        self.fluid_domain_faces = []
        self.tissue_domain_surface_meshes = []
        self.fluid_domain_surface_meshes = []
        self.tissue_domain_volume_meshes = []
        self.fluid_domain_volume_meshes = []
        self.tissue_domain_meshes = []
        self.fluid_domain_meshes = []
        self.fluid_domain_boundary_layers = []
        self.fluid_domain_interiors = []
        self.fluid_domain_wall_layers = []
        if isinstance(self.synthetic_object, svv.tree.tree.Tree):
            self.fluid_3d_simulations = [None]
            self.fluid_1d_simulations = [None]
            self.fluid_0d_simulations = [None]
            self.tissue_simulations = [None]
        elif isinstance(self.synthetic_object, svv.forest.forest.Forest):
            self.fluid_3d_simulations = [[None]*len(network) for network in self.synthetic_object.networks]
            self.fluid_1d_simulations = [[None]*len(network) for network in self.synthetic_object.networks]
            self.fluid_0d_simulations = [[None] * len(network) for network in self.synthetic_object.networks]
            self.tissue_simulations = [None]

    def build_meshes(self, fluid=True, tissue=False, hausd=0.0001, hsize=None, minratio=1.1, mindihedral=10.0,
                     order=1, remesh_vol=False, boundary_layer=True, layer_thickness_ratio=0.25,
                     layer_thickness_ratio_adjustment=0.5, boundary_layer_attempts=5, wall_layers=False,
                     wall_thickness=None, upper_num_triangles=1000, lower_num_triangles=100):
        """
        Build the mesh objects for 3D simulations.
        :return:
        [NOTE] Boolean operations and remeshing with of the interface for
        the fluid and tissue domains may need to be redone to ensure mesh
        conformation at the interface.
        """
        self.tissue_domain_surface_meshes = []
        self.fluid_domain_surface_meshes = []
        self.tissue_domain_volume_meshes = []
        self.fluid_domain_volume_meshes = []
        self.fluid_domain_boundary_layers = []
        self.fluid_domain_interiors = []
        self.fluid_domain_wall_layers = []
        if isinstance(self.synthetic_object, svv.tree.tree.Tree):
            if fluid:
                if tissue:
                    extension_scale = 4.0
                    for i in range(5):
                        new_root = self.synthetic_object.data[0, 0:3] - extension_scale * self.synthetic_object.data[0, 21]*self.synthetic_object.data.get('w_basis', 0)
                        if self.synthetic_object.domain(new_root.reshape(1, 3)) > 0:
                            break
                        else:
                            extension_scale += 1.0
                    root_extension = self.synthetic_object.data[0, 21] * extension_scale
                    self.synthetic_object.data[0, 0:3] -= root_extension * self.synthetic_object.data.get('w_basis', 0)
                fluid_surface_mesh = self.synthetic_object.export_solid(watertight=True)
                tet_fluid = tetgen.TetGen(fluid_surface_mesh)
                try:
                    #tet_fluid.tetrahedralize(minratio=minratio, mindihedral=10.0, steinerleft=-1, order=order, nobisect=True, verbose=2, switches='M')
                    tet_fluid.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                    fluid_volume_mesh = tet_fluid.grid
                except:
                    tet_fluid.make_manifold(verbose=True)
                    #tet_fluid.tetrahedralize(minratio=minratio, mindihedral=10.0, steinerleft=-1, order=order, nobisect=True, verbose=2, switches='M')
                    tet_fluid.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                    fluid_volume_mesh = tet_fluid.grid
                if isinstance(fluid_volume_mesh, type(None)):
                    print("Failed to generate fluid volume mesh.")
                else:
                    hsize = fluid_surface_mesh.hsize
                    fluid_surface_mesh = fluid_volume_mesh.extract_surface()
                    fluid_surface_faces = extract_faces(fluid_surface_mesh, fluid_volume_mesh)
                    if boundary_layer:
                        #fluid_surface_mesh = fluid_volume_mesh.extract_surface()
                        #fluid_surface_faces = extract_faces(fluid_surface_mesh, fluid_volume_mesh)
                        wall = fluid_surface_faces[1][0]
                        for i in range(boundary_layer_attempts):
                            try:
                                fluid_boundary_layers = BoundaryLayer(wall, layer_thickness=layer_thickness_ratio*hsize,
                                                                      remesh_vol=remesh_vol)
                                fluid_volume_mesh, fluid_interior, fluid_boundary = fluid_boundary_layers.generate()
                                success = True
                                print("Generated boundary layers on attempt {}/{}.".format(i+1, boundary_layer_attempts))
                            except:
                                print("Failed to generate boundary layers {}/{}.\n".format(i+1, boundary_layer_attempts))
                                fluid_boundary = None
                                fluid_interior = None
                                success = False
                                layer_thickness_ratio *= layer_thickness_ratio_adjustment
                            if success:
                                break
                        self.fluid_domain_boundary_layers.append(fluid_boundary)
                        self.fluid_domain_interiors.append(fluid_interior)
                    else:
                        if remesh_vol:
                            fluid_volume_mesh = remesh_volume(fluid_volume_mesh, hsiz=fluid_surface_mesh.hsize)
                        self.fluid_domain_boundary_layers.append(None)
                        self.fluid_domain_interiors.append(None)
                    if wall_layers:
                        if isinstance(wall_thickness, type(None)):
                            wall_thickness = 2*layer_thickness_ratio*hsize
                        wall = fluid_surface_faces[1][0]
                        fluid_boundary_layers = BoundaryLayer(wall, negate_warp_vectors=False,
                                                              layer_thickness=wall_thickness,
                                                              remesh_vol=False, combine=False)
                        _, _, fluid_wall = fluid_boundary_layers.generate()
                        # Perform tetrahedron re-orientation to ensure positive Jacobian
                        fluid_wall = remesh_volume(fluid_wall, nomove=True, noinsert=True, nosurf=True, verbosity=4)
                        if remesh_vol:
                            fluid_wall = remesh_volume(fluid_wall, hausd=hausd, nosurf=True, verbosity=4)
                        self.fluid_domain_wall_layers.append(fluid_wall)
                    fluid_surface_mesh = fluid_volume_mesh.extract_surface()
                    fluid_surface_mesh.hsize = hsize
                    self.fluid_domain_surface_meshes.append(fluid_surface_mesh)
                    self.fluid_domain_volume_meshes.append(fluid_volume_mesh)
                    if tissue:
                        self.synthetic_object.data[0, 0:3] += root_extension * self.synthetic_object.data.get('w_basis',0)
            if tissue and not isinstance(self.synthetic_object.domain, type(None)):
                # Extrude the root of the tree to ensure proper intersection with the tissue domain.
                if not fluid:
                    root_extension = max(self.synthetic_object.data[0, 21] * 4, self.synthetic_object.data[0, 20] * 0.5)
                    self.synthetic_object.data[0, 0:3] -= root_extension * self.synthetic_object.data.get('w_basis', 0)
                    # Should check to see that the extended point does not intersect with another fluid or tissue domain.
                    fluid_surface_boolean_mesh = self.synthetic_object.export_solid(watertight=True)
                else:
                    if not wall_layers:
                        fluid_surface_boolean_mesh = deepcopy(self.fluid_domain_surface_meshes[-1])
                    else:
                        fluid_surface_boolean_mesh = deepcopy(self.fluid_domain_wall_layers[-1])
                hsize = fluid_surface_boolean_mesh.hsize
                tissue_domain = remesh_surface(self.synthetic_object.domain.boundary, hausd=hausd) # Check if this should be remeshed
                area = tissue_domain.area
                tissue_domain = boolean(tissue_domain, fluid_surface_boolean_mesh, operation='difference')
                if fluid:
                    fluid_faces = extract_faces(tissue_domain, None)
                    face_sizes = [len(face) for face in fluid_faces[0]]
                    wall = numpy.argmax(face_sizes)
                    low_tri_area = area / upper_num_triangles
                    hmin = ((4.0*low_tri_area)/3.0**0.5) ** (0.5)
                    upper_tri_area = area / lower_num_triangles
                    hmax = ((4.0*upper_tri_area)/3.0**0.5) ** (0.5)
                    tissue_domain = remesh_surface(tissue_domain, hausd=hausd)
                else:
                    tissue_domain = remesh_surface(tissue_domain, hausd=hausd)
                tet_tissue = tetgen.TetGen(tissue_domain)
                if not fluid:
                    self.synthetic_object.data[0, 0:3] += root_extension * self.synthetic_object.data.get('w_basis', 0)
                try:
                    tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                    #tet_tissue.tetrahedralize(minratio=minratio, order=order)
                    tissue_volume_mesh = tet_tissue.grid
                except:
                    if fluid:
                        print('Mesh interface may be corrupted after mesh fixing for tetrahedralization.')
                    tet_tissue.make_manifold(verbose=True)
                    #tet_tissue.tetrahedralize(minratio=minratio, order=order)
                    tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                    tissue_volume_mesh = tet_tissue.grid
                if isinstance(tissue_volume_mesh, type(None)):
                    print("Failed to generate tissue volume mesh.")
                else:
                    if remesh_volume:
                        tissue_volume_mesh = remesh_volume(tissue_volume_mesh, hausd=hausd, nosurf=True)
                    tissue_domain = tissue_volume_mesh.extract_surface()
                    self.tissue_domain_surface_meshes.append(tissue_domain)
                    self.tissue_domain_volume_meshes.append(tissue_volume_mesh)
        elif isinstance(self.synthetic_object, svv.forest.forest.Forest) and isinstance(self.synthetic_object.connections, type(None)):
            for network in self.synthetic_object.networks:
                network_fluid_surface_meshes = []
                network_fluid_volume_meshes = []
                network_tissue_surface_meshes = []
                network_tissue_volume_meshes = []
                for tree in network:
                    if fluid:
                        fluid_surface_mesh = tree.export_solid(watertight=True)
                        tet_fluid = tetgen.TetGen(fluid_surface_mesh)
                        try:
                            tet_fluid.make_manifold(verbose=False)
                            tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                            fluid_volume_mesh = tet_fluid.grid
                        except:
                            try:
                                tet_fluid.make_manifold(verbose=True)
                                tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                                fluid_volume_mesh = tet_fluid.grid
                            except:
                                fluid_volume_mesh = None
                        if isinstance(fluid_volume_mesh, type(None)):
                            print("Failed to generate fluid volume mesh.")
                            network_fluid_surface_meshes.append(None)
                            network_fluid_volume_meshes.append(None)
                        else:
                            fluid_volume_mesh = remesh_volume(fluid_volume_mesh, hausd=hausd)
                            fluid_surface_mesh = fluid_volume_mesh.extract_surface()
                            network_fluid_surface_meshes.append(fluid_surface_mesh)
                            network_fluid_volume_meshes.append(fluid_volume_mesh)
                    if tissue:
                        # Extrude the root of the tree to ensure proper intersection with the tissue domain.
                        root_extension = max(tree.data[0, 21] * 4, tree.data[0, 20] * 0.5)
                        tree.data[0, 0:3] -= root_extension * tree.data.get('w_basis', 0)
                        # Should check to see that the extended point does not intersect with another fluid or tissue domain.
                        fluid_surface_boolean_mesh = tree.export_solid(watertight=True)
                        if len(self.tissue_domain_surface_meshes) > 0:
                            tissue_domain = self.tissue_domain_surface_meshes[-1]
                        else:
                            tissue_domain = tree.domain.boundary
                        tissue_domain = boolean(tissue_domain, fluid_surface_boolean_mesh, operation='difference')
                        tissue_domain = remesh_surface(tissue_domain, hausd=hausd)
                        tet_tissue = tetgen.TetGen(tissue_domain)
                        tree.data[0, 0:3] += root_extension * tree.data.get('w_basis', 0)
                        try:
                            tet_tissue.make_manifold(verbose=False)
                            tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                            tissue_volume_mesh = tet_tissue.grid
                        except:
                            try:
                                tet_tissue.make_manifold(verbose=True)
                                tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                                tissue_volume_mesh = tet_tissue.grid
                            except:
                                tissue_volume_mesh = None
                        if isinstance(tissue_volume_mesh, type(None)):
                            print("Failed to generate tissue volume mesh.")
                            network_tissue_surface_meshes.append(None)
                            network_tissue_volume_meshes.append(None)
                        else:
                            tissue_volume_mesh = remesh_volume(tissue_volume_mesh, hausd=hausd)
                            tissue_domain = tissue_volume_mesh.extract_surface()
                            network_tissue_surface_meshes.append(tissue_domain)
                            network_tissue_volume_meshes.append(tissue_volume_mesh)
                self.fluid_domain_surface_meshes.append(network_fluid_surface_meshes)
                self.fluid_domain_volume_meshes.append(network_fluid_volume_meshes)
                self.tissue_domain_surface_meshes.append(network_tissue_surface_meshes)
                self.tissue_domain_volume_meshes.append(network_tissue_volume_meshes)
        elif isinstance(self.synthetic_object, svv.forest.forest.Forest) and not isinstance(self.synthetic_object.connections, type(None)):
            if fluid or tissue:
                if tissue:
                    network_solids, _, _ = self.synthetic_object.connections.export_solid(extrude_roots=True)
                else:
                    network_solids, _, _ = self.synthetic_object.connections.export_solid(extrude_roots=False)
                for i, fluid_surface in enumerate(network_solids):
                    if fluid:
                        tet_fluid = tetgen.TetGen(fluid_surface)
                        try:
                            tet_fluid.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                            fluid_volume = tet_fluid.grid
                        except:
                            tet_fluid.make_manifold(verbose=True)
                            tet_fluid.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                            fluid_volume = tet_fluid.grid
                        if isinstance(fluid_volume, type(None)):
                            print("Failed to generate fluid volume mesh.")
                            self.fluid_domain_surface_meshes.append(fluid_surface)
                            self.fluid_domain_volume_meshes.append(None)
                        else:
                            hsize = fluid_surface.hsize
                            if (boundary_layer or wall_layers) and fluid:
                                fluid_surface = fluid_volume.extract_surface()
                                fluid_surface_faces = extract_faces(fluid_surface, fluid_volume)
                            if boundary_layer and fluid:
                                if len(fluid_surface_faces[1]) > 1:
                                    print("Boundary layer generation with more than one wall mesh is ambiguous.")
                                    print("Only the first wall mesh will be used.")
                                elif len(fluid_surface_faces[1]) == 0:
                                    print("No wall mesh found for boundary layer generation.")
                                wall = fluid_surface_faces[1][0]
                                for j in range(boundary_layer_attempts):
                                    try:
                                        fluid_boundary_layers = BoundaryLayer(wall,
                                                                              layer_thickness=layer_thickness_ratio * hsize,
                                                                              remesh_vol=remesh_vol)
                                        fluid_volume, fluid_interior, fluid_boundary = fluid_boundary_layers.generate()
                                        fluid_surface = fluid_volume.extract_surface()
                                        success = True
                                        print("Generated boundary layers on attempt {}/{}.".format(i + 1,
                                                                                                   boundary_layer_attempts))
                                    except:
                                        print("Failed to generate boundary layers {}/{}.\n".format(i + 1,
                                                                                                   boundary_layer_attempts))
                                        fluid_boundary = None
                                        fluid_interior = None
                                        success = False
                                        layer_thickness_ratio *= layer_thickness_ratio_adjustment
                                    if success:
                                        break
                                self.fluid_domain_boundary_layers.append(fluid_boundary)
                                self.fluid_domain_interiors.append(fluid_interior)
                            else:
                                if remesh_vol:
                                    fluid_volume = remesh_volume(fluid_volume, hsiz=hsize)
                                self.fluid_domain_boundary_layers.append(None)
                                self.fluid_domain_interiors.append(None)
                            if wall_layers and fluid:
                                if isinstance(wall_thickness, type(None)):
                                    wall_thickness = 2 * layer_thickness_ratio * hsize
                                wall = fluid_surface_faces[1][0]
                                fluid_boundary_layers = BoundaryLayer(wall, negate_warp_vectors=False,
                                                                      layer_thickness=wall_thickness,
                                                                      remesh_vol=False, combine=False)
                                _, _, fluid_wall = fluid_boundary_layers.generate()
                                # Perform tetrahedron re-orientation to ensure positive Jacobian
                                fluid_wall = remesh_volume(fluid_wall, nomove=True, noinsert=True, nosurf=True, verbosity=4)
                                if remesh_vol:
                                    fluid_wall = remesh_volume(fluid_wall, hausd=hausd, nosurf=True, verbosity=4)
                                self.fluid_domain_wall_layers.append(fluid_wall)
                            else:
                                self.fluid_domain_wall_layers.append(None)
                            fluid_surface.hsize = hsize
                        self.fluid_domain_surface_meshes.append(fluid_surface)
                        self.fluid_domain_volume_meshes.append(fluid_volume)
                    else:
                        self.fluid_domain_surface_meshes.append(fluid_surface)
                        self.fluid_domain_volume_meshes.append(None)
            if tissue:
                tissue_domain = deepcopy(self.synthetic_object.domain.boundary)
                tissue_domain = tissue_domain.compute_normals(auto_orient_normals=True)
                fluid_hsize = min([mesh.hsize for mesh in self.fluid_domain_surface_meshes])
                radii = []
                for net in range(len(self.synthetic_object.networks)):
                    for tr in range(len(self.synthetic_object.networks[net])):
                        radii.append(self.synthetic_object.networks[net][tr].data[0, 21])
                hsize = min(radii) * 2.0
                print("Remeshing tissue domain with edge size {}.".format(hsize))
                tissue_domain = remesh_surface(tissue_domain, hsiz=hsize)
                for i, fluid_surface in enumerate(self.fluid_domain_surface_meshes):
                    fluid_surface_normals = fluid_surface.compute_normals(auto_orient_normals=True)
                    print("Performing boolean operation with fluid surface mesh {}.".format(i))
                    tissue_domain = boolean(tissue_domain, fluid_surface_normals, operation='difference', engine='blender')
                    tissue_domain = tissue_domain.compute_normals(auto_orient_normals=True)
                    print("Remeshing tissue domain with edge size {}.".format(fluid_hsize))
                    #tissue_domain = remesh_surface(tissue_domain, hmin=fluid_hsize, hmax=hsize)
                    tissue_domain = remesh_surface(tissue_domain, optim=True)
                self.tissue_domain_surface_meshes.append(tissue_domain)
                #tissue_domain = remesh_surface(tissue_domain, hausd=hausd)
                print("Tetrahedralizing tissue domain.")
                tet_tissue = tetgen.TetGen(tissue_domain)
                try:
                    tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                    tissue_volume_mesh = tet_tissue.grid
                except:
                    tet_tissue.make_manifold(verbose=True)
                    tet_tissue.tetrahedralize(switches='pq{}/{}MVYSJ'.format(minratio, mindihedral))
                    tissue_volume_mesh = tet_tissue.grid
                if isinstance(tissue_volume_mesh, type(None)):
                    print("Failed to generate tissue volume mesh.")
                else:
                    if remesh_vol:
                        tissue_volume_mesh = remesh_volume(tissue_volume_mesh, hausd=hausd, nosurf=True)
                    tissue_surface = tissue_volume_mesh.extract_surface()
                    self.tissue_domain_surface_meshes[-1] = tissue_surface
                    self.tissue_domain_volume_meshes.append(tissue_volume_mesh)
        else:
            raise ValueError("Unsupported synthetic object type.")

    def extract_faces(self, crease_angle=60.0, verbose=False):
        """
        Extract the faces from the mesh objects.
        :return:
        """
        self.tissue_domain_faces = []
        self.fluid_domain_faces = []
        self.fluid_domain_meshes = []
        self.tissue_domain_meshes = []
        if isinstance(self.synthetic_object, svv.tree.tree.Tree):
            if len(self.fluid_domain_surface_meshes) > 0 and len(self.fluid_domain_volume_meshes) > 0:
                nodes, walls, caps, shared_boundaries = extract_faces(self.fluid_domain_surface_meshes[0],
                                                                      self.fluid_domain_volume_meshes[0],
                                                                      crease_angle=crease_angle, verbose=verbose)
                self.fluid_domain_faces.append({'walls': walls, 'caps': caps, 'shared_boundaries': shared_boundaries})
                fluid_mesh = GeneralMesh()
                fluid_mesh.add_mesh(self.fluid_domain_volume_meshes[0], name='fluid_msh_0')
                for i, wall in enumerate(walls):
                    fluid_mesh.add_face(wall, name='wall_{}'.format(i))
                for i, cap in enumerate(caps):
                    fluid_mesh.add_face(cap, name='cap_{}'.format(i))
                fluid_mesh.check_mesh()
                self.fluid_domain_meshes.append(fluid_mesh)
            if len(self.tissue_domain_surface_meshes) > 0 and len(self.tissue_domain_volume_meshes) > 0:
                nodes, walls, caps, shared_boundaries = extract_faces(self.tissue_domain_surface_meshes[0],
                                                                      self.tissue_domain_volume_meshes[0],
                                                                      crease_angle=crease_angle, verbose=verbose)
                self.tissue_domain_faces.append({'walls': walls, 'caps': caps, 'shared_boundaries': shared_boundaries})
                tissue_mesh = GeneralMesh()
                tissue_mesh.add_mesh(self.tissue_domain_volume_meshes[0], name='tissue_msh_0')
                for i, wall in enumerate(walls):
                    tissue_mesh.add_face(wall, name='lumen_{}'.format(i))
                for i, cap in enumerate(caps):
                    tissue_mesh.add_face(cap, name='wall_{}'.format(i))
                tissue_mesh.check_mesh()
                self.tissue_domain_meshes.append(tissue_mesh)
        elif isinstance(self.synthetic_object, svv.forest.forest.Forest):
            network_tissue_faces = []
            network_tissue_domains = []
            network_fluid_faces = []
            network_fluid_domains = []
            for i in range(len(self.fluid_domain_surface_meshes)):
                for j in range(len(self.fluid_domain_surface_meshes[i])):
                    surface = self.fluid_domain_surface_meshes[i][j]
                    mesh = self.fluid_domain_volume_meshes[i][j]
                    if isinstance(surface, type(None)) or isinstance(mesh, type(None)):
                        network_fluid_faces.append(None)
                        network_fluid_domains.append(None)
                        continue
                    nodes, walls, caps, shared_boundaries = extract_faces(surface, mesh, crease_angle=crease_angle,
                                                                          verbose=verbose)
                    network_fluid_faces.append({'walls': walls, 'caps': caps, 'shared_boundaries': shared_boundaries})
                    fluid_mesh = GeneralMesh()
                    fluid_mesh.add_mesh(mesh, name='fluid_msh_{}'.format(len(self.fluid_domain_meshes)))
                    for k, wall in enumerate(walls):
                        fluid_mesh.add_face(wall, name='wall_{}'.format(k))
                    for k, cap in enumerate(caps):
                        fluid_mesh.add_face(cap, name='cap_{}'.format(k))
                    fluid_mesh.check_mesh()
                    network_fluid_domains.append(fluid_mesh)
                self.fluid_domain_faces.append(network_fluid_faces)
                self.fluid_domain_meshes.append(network_fluid_domains)
            for i in range(len(self.tissue_domain_surface_meshes)):
                for j in range(len(self.tissue_domain_surface_meshes[i])):
                    surface = self.tissue_domain_surface_meshes[i][j]
                    mesh = self.tissue_domain_volume_meshes[i][j]
                    nodes, walls, caps, shared_boundaries = extract_faces(surface, mesh, crease_angle=crease_angle,
                                                                          verbose=False)
                    network_tissue_faces.append({'walls': walls, 'caps': caps, 'shared_boundaries': shared_boundaries})
                    tissue_mesh = GeneralMesh()
                    tissue_mesh.add_mesh(mesh, name='tissue_msh_{}'.format(len(self.tissue_domain_meshes)))
                    for k, wall in enumerate(walls):
                        tissue_mesh.add_face(wall, name='lumen_{}'.format(k))
                    for k, cap in enumerate(caps):
                        tissue_mesh.add_face(cap, name='wall_{}'.format(k))
                    tissue_mesh.check_mesh()
                    network_tissue_domains.append(tissue_mesh)
                self.tissue_domain_faces.append(network_tissue_faces)
                self.tissue_domain_meshes.append(network_tissue_domains)

    def construct_3d_fluid_equation(self, *args):
        """
        Construct the equations for the simulation.
        :return:
        """
        if isinstance(self.synthetic_object, svv.tree.tree.Tree):
            fluid_mesh = self.fluid_domain_meshes[0]
            fluid_equation = FluidEquation()
            fluid_equation.add_mesh(fluid_mesh)
            inlet = None
            best = numpy.inf
            inlet_center = self.synthetic_object.data[0, 0:3]
            for name, face in fluid_mesh.faces.items():
                if 'cap' in name:
                    if numpy.linalg.norm(face.center - inlet_center) < best:
                        best = numpy.linalg.norm(face.center)
                        inlet = name
            if isinstance(inlet, type(None)):
                raise ValueError("Inlet not found.")
            fluid_equation.add_inlet(inlet, -1*self.synthetic_object.data[0, 22])
            fluid_equation.set_viscosity('Constant', self.synthetic_object.parameters.kinematic_viscosity)
        elif isinstance(self.synthetic_object, svv.forest.forest.Forest):
            if len(args) == 0:
                network_id = 0
                tree_id = 0
            elif len(args) == 1:
                network_id = args[0]
                tree_id = 0
            elif len(args) == 2:
                network_id = args[0]
                tree_id = args[1]
            else:
                raise ValueError("Too many arguments.")
            fluid_mesh = self.fluid_domain_meshes[network_id][tree_id]
            fluid_equation = FluidEquation()
            fluid_equation.add_mesh(fluid_mesh)
            # Verify cap inlet
            inlet = None
            for name, face in fluid_mesh.faces.items():
                best = numpy.inf
                inlet_center = self.synthetic_object.networks[network_id][tree_id].data[0, 0:3]
                if 'cap' in name:
                    if numpy.linalg.norm(face.center - inlet_center) < best:
                        best = numpy.linalg.norm(face.center)
                        inlet = name
            if isinstance(inlet, type(None)):
                raise ValueError("Inlet not found.")
            fluid_equation.add_inlet(inlet, -1*self.synthetic_object.networks[network_id][tree_id].data[0, 22])
            fluid_equation.set_viscosity('Constant', self.synthetic_object.networks[network_id][tree_id].parameters.kinematic_viscosity)
        else:
            raise ValueError("Unsupported synthetic object type.")
        for face in fluid_mesh.faces:
            if face == inlet:
                continue
            if 'cap' in face:
                fluid_equation.add_outlet(face, value=0.0)
            if 'wall' in face:
                fluid_equation.add_wall(face)
        fluid_equation.check_bcs()
        return fluid_equation

    def construct_3d_fluid_simulation(self, *args):
        """
        Construct the 3D simulations.
        :return:
        """
        if len(args) == 0:
            network_id = 0
            tree_id = 0
        elif len(args) == 1:
            network_id = args[0]
            tree_id = 0
        elif len(args) == 2:
            network_id = args[0]
            tree_id = args[1]
        else:
            raise ValueError("Too many arguments.")
        if isinstance(self.synthetic_object, svv.tree.tree.Tree):
            simulation_file = minidom.Document()
            svfsi_file = simulation_file.createElement("svFSIFile")
            svfsi_file.setAttribute("version", "0.1")
            general_simulation_parameters = GeneralSimulationParameters()
            fluid_mesh = self.fluid_domain_meshes[0]
            fluid_equation = self.construct_3d_fluid_equation()
            svfsi_file.appendChild(general_simulation_parameters.toxml())
            svfsi_file.appendChild(fluid_mesh.toxml())
            svfsi_file.appendChild(fluid_equation.toxml())
            simulation_file.appendChild(svfsi_file)
            self.fluid_3d_simulations[0] = tuple([simulation_file, fluid_mesh])
        elif isinstance(self.synthetic_object, svv.forest.forest.Forest):
            simulation_file = minidom.Document()
            svfsi_file = simulation_file.createElement("svFSIFile")
            svfsi_file.setAttribute("version", "0.1")
            general_simulation_parameters = GeneralSimulationParameters()
            fluid_mesh = self.fluid_domain_meshes[network_id][tree_id]
            fluid_equation = self.construct_3d_fluid_equation(args)
            svfsi_file.appendChild(general_simulation_parameters.toxml())
            svfsi_file.appendChild(fluid_mesh.toxml())
            svfsi_file.appendChild(fluid_equation.toxml())
            simulation_file.appendChild(svfsi_file)
            self.fluid_3d_simulations[network_id][tree_id] = tuple([simulation_file, fluid_mesh])
        else:
            raise ValueError("Index out of range.")
        return

    def write_3d_fluid_simulation(self, *args):
        """
        Write the fluid simulation to disk.
        :return:
        """
        if len(args) == 0:
            network_id = 0
            tree_id = 0
        elif len(args) == 1:
            network_id = args[0]
            tree_id = 0
        elif len(args) == 2:
            network_id = args[0]
            tree_id = args[1]
        else:
            raise ValueError("Too many arguments.")
        if isinstance(self.synthetic_object, svv.tree.tree.Tree):
            simulation_file, fluid_mesh = self.fluid_simulations[0]
        elif isinstance(self.synthetic_object, svv.forest.forest.Forest):
            simulation_file, fluid_mesh = self.fluid_simulations[network_id][tree_id]
        else:
            raise ValueError("Index out of range.")
        if not isinstance(simulation_file, type(None)) and not isinstance(fluid_mesh, type(None)):
            if not os.path.exists(self.file_path):
                os.mkdir(self.file_path)
                if not os.path.exists(self.file_path + os.sep + "mesh"):
                    os.mkdir(self.file_path + os.sep + "mesh")
                if not os.path.exists(self.file_path + os.sep + "mesh" + os.sep + fluid_mesh.name):
                    os.mkdir(self.file_path + os.sep + "mesh" + os.sep + fluid_mesh.name)
                if not os.path.exists(self.file_path + os.sep + "mesh" + os.sep + fluid_mesh.name + os.sep + "mesh-surfaces"):
                    os.mkdir(self.file_path + os.sep + "mesh" + os.sep + fluid_mesh.name + os.sep + "mesh-surfaces")
            if isinstance(fluid_mesh.mesh, pyvista.UnstructuredGrid):
                fluid_mesh.mesh.save(self.file_path + os.sep + "mesh" + os.sep + fluid_mesh.name + os.sep + "{}.vtu".format(fluid_mesh.name))
            elif isinstance(fluid_mesh.mesh, pyvista.PolyData):
                fluid_mesh.mesh.save(self.file_path + os.sep + "mesh" + os.sep + fluid_mesh.name + os.sep + "{}.vtp".format(fluid_mesh.name))
            else:
                raise ValueError("Mesh must be a pyvista mesh object.")
            for name, face in fluid_mesh.faces.items():
                if isinstance(face, pyvista.PolyData):
                    face.save(self.file_path + os.sep + "mesh" + os.sep + fluid_mesh.name + os.sep + "mesh-surfaces" + os.sep + "{}.vtp".format(name))
                else:
                    raise ValueError("Face must be a pyvista mesh object.")
            file_name = self.file_path + os.sep + "fluid_simulation_{}-{}.xml".format(network_id, tree_id)
            with open(file_name, 'w') as f:
                f.write(simulation_file.toprettyxml())
        else:
            raise ValueError("Simulation file or mesh object not found.")

    def construct_1d_fluid_simulation(self, *args, viscosity=None, density=None, time_step_size=0.01,
                                      number_time_steps=100, olufsen_material_exponent=2):
        if len(args) == 0:
            network_id = 0
            tree_id = 0
        elif len(args) == 1:
            network_id = args[0]
            tree_id = 0
        elif len(args) == 2:
            network_id = args[0]
            tree_id = args[1]
        else:
            raise ValueError("Too many positional input arguments")
        if isinstance(self.synthetic_object, svv.tree.tree.Tree):
            centerlines, _ = self.synthetic_object.export_centerlines()
            material = one_d.parameters.MaterialModel()
            params = one_d.parameters.Parameters()
            params.output_directory = self.file_path + os.sep + "fluid" + os.sep + "1d"
            params.solver_output_file = self.file_path + os.sep + "fluid" + os.sep + "1d" + os.sep + "1d_simulation_input.json"
            params.centerlines_input_file = self.file_path + os.sep + "fluid" + os.sep + "1d" + os.sep + "centerlines.vtp"
            params.outlet_face_names_file = self.file_path + os.sep + "fluid" + os.sep + "1d" + "outlets"
            params.seg_size_adaptive = True
            params.model_order = 1  # Since this is strictly a 1d ROM simulation it is not an exposed parameter
            params.uniform_bc = False
            params.inflow_input_file = self.file_path + os.sep + "fluid" + os.sep + "1d" + "inflow_1d.flow"
            params.outflow_bc_type = ["rcrt.dat"]
            params.outflow_bc_file = self.file_path + os.sep + "fluid" + os.sep + "1d"
            params.model_name = "1d_model_{}-{}".format(network_id, tree_id)
            params.compute_mesh = True
            params.time_step = time_step_size
            params.num_time_steps = number_time_steps
            params.olufsen_material_exponent = olufsen_material_exponent
            params.material_model = material.OLUFSEN
            params.viscosity = viscosity
            params.density = density
            mesh = one_d.mesh.Mesh()
            self.fluid_1d_simulations[0] = tuple([centerlines, mesh, params])
        else:
            raise ValueError("Index out of range.")

    def write_1d_fluid_simulation(self, *args):
        pass

    def construct_0d_fluid_equation(self, *args):
        pass

    def construct_0d_fluid_simulation(self, *args):
        pass

    def write_0d_fluid_simulation(self, *args):
        pass

    def construct_3d_tissue_perfusion_equation(self, *args):
        pass

    def construct_3d_tissue_perfusion_simulation(self, *args):
        pass

    def write_3d_tissue_perfusion_simulation(self, *args):
        pass

    def pulsatile_waveform(self, *args):
        """
        Generate a pulsatile waveform for fluid simulations. This function should accept either an average flow rate
        or an array of flow rate values over the nominal cardiac cycle t -> [0, 1].

        This function should check that the waveform is properly formatted and that the values are within a reasonable
        range; otherwise it should warn the user that values of the waveform might result in unrealistic/erroneous
        results. (e.g. high reynolds number flows, reversed flow, etc.)
        :param args:
        :return:
        """
        pass

    def generate_inflow(self, *args, filename=None):
        pass

    def write(self):
        """
        Save the simulation file to disk.
        :return:
        """
        os.mkdir(self.file_path)
        mesh_folder = self.file_path + os.sep + "mesh"
        os.mkdir(mesh_folder)
        surface_folder = mesh_folder + os.sep + "mesh-surfaces"
        os.mkdir(surface_folder)

    def write_input_file(self):
        """
        Write the simulation input file.
        :return:
        """
        pass



import numpy as np


def h_matrix(a_, n, d, lam):
    r"""
    Compute the H matrix for Duchon's variational problem.

    This function computes the H matrix, also known as Duchon's matrix, which forms the linear system
    of equations to minimize in the variational problem of thin-plate spline interpolation. The matrix
    :math:`H` is used to represent the bending energy minimization problem and should be positive
    semi-definite.

    Parameters
    ----------
    a_ : ndarray of shape ((n + 1) * (d + 1), (n + 1) * (d + 1))
        The full interpolation matrix :math:`A` for the points defining the (D-1)-dimensional manifold
        embedded in D-dimensional space. It is assembled from sub-matrices M and N.

    n : int
        The number of points defining the (D-1)-dimensional manifold.

    d : int
        The dimension of the space in which the (D-1)-dimensional manifold is embedded.

    lam : float
        The regularization parameter (also known as the thin-plate relaxation parameter).
        It determines the trade-off between fitting the manifold through the control points
        and smoothing the interpolation. A higher value of :math:`\lambda` leads to a smoother
        manifold that may not pass exactly through the control points.

    Returns
    -------
    h_ : ndarray of shape ((d * n), (d * n))
        The H matrix representing the linear system for the bending energy minimization problem.

    j00_ : ndarray
        The top-left block of the inverse of :math:`A`, corresponding to the data-data interactions.

    j01_ : ndarray
        The top-right block of the inverse of :math:`A`, corresponding to the data-polynomial interactions.

    j11_ : ndarray
        The bottom-right block of the inverse of :math:`A`, corresponding to the polynomial-polynomial interactions.

    a_inv : ndarray
        The inverse of the interpolation matrix :math:`A`.

    Notes
    -----
    The function computes the H matrix using the following steps:

    1. Compute the inverse of the interpolation matrix :math:`A`:

       .. math::
           A^{-1}

    2. Partition the inverse matrix into blocks:

       - :math:`J = A^{-1}_{0: n \times (d+1), 0: n \times (d+1)}`
       - :math:`J_{00} = J_{0: n \times d, 0: n \times d}`
       - :math:`J_{01} = J_{0: n \times d, n \times d: n \times (d+1)}`
       - :math:`J_{11} = J_{n \times d: n \times (d+1), n \times d: n \times (d+1)}`

    3. Compute the H matrix based on the regularization parameter :math:`\lambda`:

       - If :math:`\lambda > 0`:

         .. math::
             H = J_{11} - \lambda J_{01}^T \left( I + \lambda J_{00} \right)^{-1} J_{01}

       - If :math:`\lambda = 0`:

         .. math::
             H = J_{11}

    The H matrix is positive semi-definite and is used in the bending energy minimization problem for thin-plate
    splines.

    References
    ----------
    .. [1] Duchon, J. (1977). Splines minimizing rotation-invariant semi-norms in Sobolev spaces.
           In *Constructive Theory of Functions of Several Variables* (pp. 85-100). Springer, Berlin, Heidelberg.
           `Springer Link <https://doi.org/10.1007/BFb0086566>`_

    Examples
    --------
    **Example 1: Computing the H Matrix with Regularization**

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.core.a_matrix import a_matrix
        from svtoolkit.domain.core.h_matrix import h_matrix

        # Define parameters
        n = 5       # Number of points
        d = 2       # Dimension (e.g., 2D space)
        lam = 0.1   # Regularization parameter

        # Generate sample points
        points = np.random.rand(n, d)

        # Compute the interpolation matrix A
        A = a_matrix(points)

        # Compute the H matrix
        h, j00, j01, j11, a_inv = h_matrix(A, n, d, lam)

        print("H matrix shape:", h.shape)
        # Output: H matrix shape: (10, 10)

    **Example 2: Without Regularization**

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.core.a_matrix import a_matrix
        from svtoolkit.domain.core.h_matrix import h_matrix

        # Define parameters
        n = 4       # Number of points
        d = 3       # Dimension (e.g., 3D space)
        lam = 0.0   # No regularization

        # Generate sample points
        points = np.random.rand(n, d)

        # Compute the interpolation matrix A
        A = a_matrix(points)

        # Compute the H matrix
        h, j00, j01, j11, a_inv = h_matrix(A, n, d, lam)

        print("H matrix shape:", h.shape)
        # Output: H matrix shape: (12, 12)

    **Example 3: Solving the Bending Energy Minimization Problem**

    .. code-block:: python

        import numpy as np
        from svtoolkit.domain.core.a_matrix import a_matrix
        from svtoolkit.domain.core.h_matrix import h_matrix
        from scipy.linalg import eigh

        # Define parameters
        n = 6
        d = 2
        lam = 0.05

        # Generate sample points
        points = np.random.rand(n, d)

        # Compute the interpolation matrix A
        A = a_matrix(points)

        # Compute the H matrix
        h, _, _, _, _ = h_matrix(A, n, d, lam)

        # Solve the eigenvalue problem
        eigenvalues, eigenvectors = eigh(h)

        print("Eigenvalues of H matrix:", eigenvalues)

    See Also
    --------
    :func:`svtoolkit.domain.core.a_matrix.a_matrix` : Function to compute the full interpolation matrix A.
    :func:`svtoolkit.domain.core.m_matrix.m_matrix` : Function to compute the M sub-matrix.
    :func:`svtoolkit.domain.core.n_matrix.n_matrix` : Function to compute the N sub-matrix.

    """
    a_inv = np.linalg.inv(a_)
    j_ = a_inv[:n * (d + 1), :n * (d + 1)]
    j00_ = j_[:n, :n]
    j01_ = j_[:n, n:].copy()
    j11_ = j_[n:, n:]
    if lam > 0:
        inv = np.linalg.inv(np.eye(j00_.shape[0]) + lam * j00_)
        h_ = j11_ - (lam * j01_.T) @ inv @ j01_
    else:
        h_ = j11_
    return h_, j00_, j01_, j11_, a_inv



import pytest
import numpy as np

from svv.domain.routines.allocate import allocate

@pytest.fixture
def simple_points():
    """
    A small set of 2D or 3D points.
    We’ll just do a simple 2D example for convenience
    that can be extended to 3D.
    """
    # Example: four points forming a square in 2D (z=0)
    pts = np.array([
        [0.0, 0.0],
        [1.0, 0.0],
        [1.0, 1.0],
        [0.0, 1.0],
    ])
    return pts


@pytest.fixture
def simple_points_and_normals(simple_points):
    """
    Provide a matching array of normals for each point.
    """
    # Let’s just say each normal is pointing 'up' in z direction in 3D
    # or if using 2D arrays, we can store Nx2 for 2D "normals".
    # For demonstration, let's assume 3D with z=1 for all:
    pts = simple_points
    # Convert to 3D by appending z=0
    pts_3d = np.hstack([pts, np.zeros((pts.shape[0], 1))])
    # Normals all point in +z
    normals_3d = np.zeros_like(pts_3d)
    normals_3d[:, 2] = 1.0
    return pts_3d, normals_3d


def test_allocate_no_data():
    """
    Test calling allocate with no arguments,
    expecting it to return [None, None] and print an error message.
    """
    result = allocate()
    assert result == [None, None], "allocate() with no data should return [None, None]"


def test_allocate_only_points(simple_points):
    """
    Test calling allocate with only points, no normals.
    """
    patches = allocate(simple_points)
    # Check if we get something valid back
    assert isinstance(patches, list), "Expected a list of patches."
    # Each element of patches is (patch_points, patch_normals)
    for p_pts, p_normals in patches:
        assert p_normals is None, "No normals given => patch_normals should be None."
        assert p_pts.shape[1] == simple_points.shape[1], "Patch points should match dimension of input."


def test_allocate_points_and_normals(simple_points_and_normals):
    """
    Test calling allocate with points and normals.
    """
    pts, normals = simple_points_and_normals
    patches = allocate(pts, normals)
    assert isinstance(patches, list), "Expected a list of patches."
    for p_pts, p_normals in patches:
        # Each patch’s dimension should be Nx3
        # (since we used 3D points and normals).
        if p_pts is not None:
            assert p_pts.shape[1] == 3, "Patch points should have 3D coords."
        if p_normals is not None:
            assert p_normals.shape[1] == 3, "Patch normals should have 3D coords."


def test_allocate_zero_magnitude_normals():
    """
    Provide a few points with zero-magnitude normals to ensure
    the function logs an error and removes them.
    """
    pts = np.array([
        [0,0,0],
        [1,0,0],
        [2,0,0],
    ], dtype=float)
    # Normals: second is zero
    normals = np.array([
        [0,0,1],
        [0,0,0],  # zero magnitude
        [0,0,1],
    ], dtype=float)
    patches = allocate(pts, normals)
    # The function is supposed to remove the point with zero normal
    # so only 2 points remain
    # Typically the function prints a message about zero magnitude normals
    # We can check the patch data:
    all_patch_points = np.vstack([pp for pp,_ in patches])
    # Because patches may overlap
    unique_pts_in_patches = np.unique(all_patch_points, axis=0)
    assert len(unique_pts_in_patches) == 2, (
        "Points with zero normals should have been removed, leaving 2 points."
    )


def test_allocate_duplicates():
    """
    Test handling of exact duplicate points.
    If duplicates exist, the code might remove or
    handle them in a special way.
    """
    pts = np.array([
        [0,0,0],
        [1,1,1],
        [1,1,1],  # duplicate
        [2,2,2],
    ], dtype=float)
    # Let’s say all normals are (0,0,1) to keep it simple
    normals = np.array([
        [0,0,1],
        [0,0,1],
        [0,0,1],
        [0,0,1],
    ], dtype=float)
    patches = allocate(pts, normals)
    # Check that it runs and returns patches
    assert isinstance(patches, list), "Expected a list of patches even with duplicates."

    # See how many total points remain in the patches
    # The code might remove duplicates if they have identical normals
    # or handle them in a certain way.
    # For a simple check, we’ll ensure we have <= 4 points total
    all_pts_in_patches = np.vstack([pp for pp,_ in patches])
    unique_pts_in_patches = np.unique(all_pts_in_patches, axis=0)
    # The function may remove or keep duplicates depending on logic.
    # If it removes duplicates, we expect 3 or fewer. If not, we might see 4.
    # We can't say exactly unless we know the intended logic.
    # But let's ensure it doesn't crash.
    assert len(unique_pts_in_patches) <= 4, "No crash or unexpected addition of points."


def test_allocate_overlap_and_angle(simple_points_and_normals):
    """
    Test with custom overlap and feature_angle settings.
    Ensures function can handle them without error.
    """
    pts, normals = simple_points_and_normals
    patches = allocate(pts, normals, min_patch_size=2, max_patch_size=4,
                       overlap=0.5, feature_angle=45)
    # We just test that it doesn’t crash and returns a list
    assert isinstance(patches, list)
    # Optionally, we can check how many patches we get,
    # or that each patch has at least 2 points, etc.
    for p_pts, p_norms in patches:
        if p_pts is not None:
            assert len(p_pts) >= 2, "min_patch_size=2 => each patch should have >=2 points."


def test_allocate_large_data():
    """
    Stress test with a larger random point cloud
    (without normals) to ensure performance logic at least runs.
    """
    np.random.seed(42)
    pts = np.random.rand(500, 3)  # 500 random points in 3D
    patches = allocate(pts, min_patch_size=5, max_patch_size=20, overlap=0.1)
    assert isinstance(patches, list), "Expected a list of patches."
    # Just ensure it doesn't fail for bigger data.


def test_allocate_invalid_overlap(simple_points):
    """
    Overlap outside [0,1] gets clipped.
    Ensure the function doesn't crash or do something weird.
    """
    # Overlap = 2 => clipped to 1
    patches = allocate(simple_points, min_patch_size=2, max_patch_size=3, overlap=2)
    assert isinstance(patches, list), "Should still return a list."
    # Overlap = -0.5 => clipped to 0
    patches = allocate(simple_points, min_patch_size=2, max_patch_size=3, overlap=-0.5)
    assert isinstance(patches, list), "Should still return a list."



import pytest
import numpy as np
from svv.domain.core.m_matrix import m00, m01, m11, m_matrix


def test_m00_basic():
    """
    Test m00 function for basic 2D points.
    """
    points = np.array([[0, 0], [1, 0], [0, 1]])
    rbf_degree = 3
    result = m00(points, rbf_degree)
    expected_shape = (3, 3)  # N x N matrix
    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"
    # Verify diagonal elements (distance to self is zero)
    assert np.allclose(np.diag(result), 0.0), "Diagonal elements of m00 should be zero."


def test_m01_basic():
    """
    Test m01 function for basic 2D points.
    """
    points = np.array([[0, 0], [1, 0], [0, 1]])
    rbf_degree = 3
    result = m01(points, rbf_degree)

    # Verify shape
    expected_shape = (3, 6)  # N x (D*N) matrix
    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"

    # Verify self-interaction elements are zero
    n, d = points.shape
    for i in range(n):
        for k in range(d):
            assert result[i, i * d + k] == 0.0, f"Element {i, i * d + k} (self-interaction) should be zero."

    # Verify specific interactions between points
    # For i=0, j=1 (point [0, 0] and [1, 0]), M[0, 1D + k] = -p * (x_i[k] - x_j[k]) * ||x_i - x_j||^{p-2}
    dist_01 = np.linalg.norm(points[0] - points[1])  # ||x_i - x_j||
    for k in range(d):
        expected = -rbf_degree * (points[0, k] - points[1, k]) * (dist_01 ** (rbf_degree - 2))
        assert np.isclose(result[0, 1 * d + k], expected), (
            f"Element {0, 1 * d + k} should be {expected}, got {result[0, 1 * d + k]}"
        )


def test_m11_basic():
    """
    Test m11 function for basic 2D points.
    """
    points = np.array([[0, 0], [1, 0], [0, 1]])
    rbf_degree = 3
    result = m11(points, rbf_degree)  # By default, m11 only returns the upper triangular part
    result += result.T

    # Verify shape
    expected_shape = (6, 6)  # (D*N) x (D*N) matrix
    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"

    # Verify symmetry
    assert np.allclose(result, result.T), "m11 matrix should be symmetric."

    # Verify self-interaction elements
    n, d = points.shape
    for i in range(n):
        for k in range(d):
            idx = i * d + k
            assert result[idx, idx] == 0.0, f"Self-interaction element {idx, idx} should be zero."

    # Verify specific interactions between points
    # For i=0, j=1 (points [0,0] and [1,0])
    dist_01 = np.linalg.norm(points[0] - points[1])  # ||x_i - x_j||
    for k in range(d):
        for l in range(d):
            idx_i = 0 * d + k
            idx_j = 1 * d + l
            if k == l:
                expected = (
                    -2 * (rbf_degree / 2 - 1) * rbf_degree * (points[0, k] - points[1, k]) ** 2 * (dist_01 **
                                                                                                   (rbf_degree - 4))
                    - rbf_degree * (dist_01 ** (rbf_degree - 2))
                )
            else:
                expected = (
                    -2 * (rbf_degree / 2 - 1) * rbf_degree * (points[0, k] - points[1, k]) * (points[0, l] -
                                                                                              points[1, l]) *
                    (dist_01 ** (rbf_degree - 4))
                )
            assert np.isclose(result[idx_i, idx_j], expected), (
                f"Element {idx_i, idx_j} should be {expected}, got {result[idx_i, idx_j]}"
            )


def test_m_matrix_basic():
    """
    Test m_matrix function for basic 2D points.
    """
    points = np.array([[0, 0], [1, 0], [0, 1]])
    rbf_degree = 3
    result = m_matrix(points, rbf_degree)
    n = points.shape[0]
    d = points.shape[1]
    expected_shape = (n * (d + 1), n * (d + 1))  # N*(D+1) x N*(D+1) matrix
    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"


def test_m00_edge_case_empty_points():
    """
    Test m00 function for an empty input.
    """
    points = np.empty((0, 2))
    rbf_degree = 3
    result = m00(points, rbf_degree)
    assert result.shape == (0, 0), "m00 with empty input should return an empty matrix."


def test_m01_edge_case_empty_points():
    """
    Test m01 function for an empty input.
    """
    points = np.empty((0, 2))
    rbf_degree = 3
    result = m01(points, rbf_degree)
    assert result.shape == (0, 0), "m01 with empty input should return an empty matrix."


def test_m11_edge_case_empty_points():
    """
    Test m11 function for an empty input.
    """
    points = np.empty((0, 2))
    rbf_degree = 3
    result = m11(points, rbf_degree)
    assert result.shape == (0, 0), "m11 with empty input should return an empty matrix."


def test_m_matrix_edge_case_empty_points():
    """
    Test m_matrix function for an empty input.
    """
    points = np.empty((0, 2))
    rbf_degree = 3
    result = m_matrix(points, rbf_degree)
    assert result.shape == (0, 0), "m_matrix with empty input should return an empty matrix."


def test_m_matrix_high_dimensional_points():
    """
    Test m_matrix function for 3D points.
    """
    points = np.random.rand(4, 3)  # 4 points in 3D
    rbf_degree = 3
    result = m_matrix(points, rbf_degree)
    n = points.shape[0]
    d = points.shape[1]
    expected_shape = (n * (d + 1), n * (d + 1))  # N*(D+1) x N*(D+1) matrix
    assert result.shape == expected_shape, f"Expected shape {expected_shape}, got {result.shape}"


def test_m_matrix_symmetry():
    """
    Verify symmetry of the full M matrix.
    """
    points = np.random.rand(4, 2)  # 4 points in 2D
    rbf_degree = 3
    result = m_matrix(points, rbf_degree)
    assert np.allclose(result, result.T), "Full M matrix should be symmetric."



from xml.dom import minidom


class BoundaryConditionBase:
    def __init__(self):
        self.name = None
        self.bc_type = None
        self.value = None
        self.time_varying = False
        self.time_varying_file = None
        self.impose_flux = False
        self.profile = None
        self.zero_out_perimeter = False

    def __str__(self):
        return str(vars(self))

    def __repr__(self):
        return str(vars(self))

    def __eq__(self, other):
        check = ["bc_type", "value", "time_varying", "time_varying_file", "impose_flux", "profile"]
        attributes = vars(self)
        other_attributes = vars(other)
        return all(other_attributes[key] == attributes[key] for key in check)

    def __ne__(self, other):
        return not self.__eq__(other)


class BoundaryCondition(BoundaryConditionBase):
    def __init__(self):
        super().__init__()
        self.file = minidom.Document()

    def __str__(self):
        return self.toxml().toprettyxml()

    def __repr__(self):
        return self.toxml().toprettyxml()

    def set_type(self, bc_type):
        if bc_type not in ["Dirichlet", "Neumann", "Dir", "Neu"]:
            raise ValueError("Boundary condition type must be 'Dirichlet' or 'Neumann'.")
        self.bc_type = bc_type

    def set_name(self, name):
        if not isinstance(name, str):
            raise ValueError("Name must be a string.")
        self.name = name

    def set_value(self, value):
        if not isinstance(value, float) and not isinstance(value, int):
            raise ValueError("Value must be a float of int.")
        self.value = value

    def set_time_varying(self, time_varying):
        if not isinstance(time_varying, bool):
            raise ValueError("Time varying must be a boolean.")
        self.time_varying = time_varying

    def set_time_varying_file(self, time_varying_file):
        if not isinstance(time_varying_file, str):
            raise ValueError("Time varying file must be a string.")
        self.time_varying_file = time_varying_file

    def set_impose_flux(self, impose_flux):
        if not isinstance(impose_flux, bool):
            raise ValueError("Imposed flux must be a boolean.")
        self.impose_flux = impose_flux

    def set_profile(self, profile):
        if not isinstance(profile, str):
            raise ValueError("Profile must be a string.")
        if profile not in ["Parabolic", "Plug"]:
            raise ValueError("Profile must be 'Parabolic' or 'Plug' flow.")
        self.profile = profile

    def set_zero_out_perimeter(self, zero_out_perimeter):
        if not isinstance(zero_out_perimeter, bool):
            raise ValueError("Zero out perimeter must be a boolean.")
        self.zero_out_perimeter = zero_out_perimeter

    def toxml(self):
        add_bc = self.file.createElement("Add_BC")
        if not isinstance(self.name, type(None)):
            add_bc.setAttribute("name", self.name)

        if not isinstance(self.bc_type, type(None)):
            type_element = minidom.Element("Type")
            type_element.appendChild(self.file.createTextNode(str(self.bc_type)))
            add_bc.appendChild(type_element)

        if not isinstance(self.value, type(None)):
            value_element = minidom.Element("Value")
            value_element.appendChild(self.file.createTextNode(str(self.value)))
            add_bc.appendChild(value_element)

        time_dependence_element = minidom.Element("Time_dependence")

        if self.time_varying:
            time_dependence_element.appendChild(self.file.createTextNode("Unsteady"))
            add_bc.appendChild(time_dependence_element)
            time_varying_file_element = minidom.Element("Temporal_values_file_path")
            time_varying_file_element.appendChild(self.file.createTextNode(self.time_varying_file))
            add_bc.appendChild(time_varying_file_element)
        else:
            time_dependence_element.appendChild(self.file.createTextNode("Steady"))

        if self.impose_flux:
            impose_flux_element = minidom.Element("Impose_flux")
            impose_flux_element.appendChild(self.file.createTextNode("true"))
            add_bc.appendChild(impose_flux_element)

        if not isinstance(self.profile, type(None)):
            profile_element = minidom.Element("Profile")
            profile_element.appendChild(self.file.createTextNode(str(self.profile)))
            add_bc.appendChild(profile_element)

        zero_out_perimeter_element = minidom.Element("Zero_out_perimeter")
        if self.zero_out_perimeter:
            zero_out_perimeter_element.appendChild(self.file.createTextNode("true"))
            add_bc.appendChild(zero_out_perimeter_element)
        else:
            zero_out_perimeter_element.appendChild(self.file.createTextNode("false"))
            add_bc.appendChild(zero_out_perimeter_element)
        return add_bc



import pytest
import numpy as np
from svv.forest.connect.bezier import BezierCurve
from svv.forest.connect.catmullrom import CatmullRomCurve
from svv.forest.connect.nurbs import NURBSCurve
from svv.forest.connect.curve import Curve


# Test Case 1: Default curve type (Bezier)
def test_default_curve_type():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points)

    # Check that the correct type is instantiated
    assert isinstance(curve.curve, BezierCurve)
    assert curve.curve.control_points.shape == (3, 2)  # 3 control points, 2D


# Test Case 2: CatmullRom curve type
def test_catmull_rom_curve():
    control_points = [(0, 0), (1, 2), (2, 3), (4, 4)]
    curve = Curve(control_points, curve_type="CatmullRom")

    # Check that the correct type is instantiated
    assert isinstance(curve.curve, CatmullRomCurve)
    assert curve.curve.control_points.shape == (4, 2)  # 4 control points, 2D


# Test Case 3: NURBS curve type with additional parameters
def test_nurbs_curve():
    control_points = [(0, 0), (1, 2), (3, 4)]
    knots = [0, 0, 1, 1]
    weights = [1, 1, 1]
    curve = Curve(control_points, curve_type="NURBS", knots=knots, weights=weights)

    # Check that the correct type is instantiated
    assert isinstance(curve.curve, NURBSCurve)
    assert curve.curve.control_points.shape == (3, 2)  # 3 control points, 2D
    assert curve.curve.knots == knots
    assert curve.curve.weights == weights


# Test Case 4: Unsupported curve type
def test_unsupported_curve_type():
    control_points = [(0, 0), (1, 2), (3, 4)]

    with pytest.raises(ValueError, match="Unsupported curve type"):
        Curve(control_points, curve_type="Unsupported")


# Test Case 5: Evaluate method
def test_evaluate_method():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points, curve_type="Bezier")
    t_values = [0.0, 0.5, 1.0]

    # Evaluate the curve at the specified parametric values
    evaluated_points = curve.evaluate(t_values)

    # Check the returned shape (it should be len(t_values) x d)
    assert evaluated_points.shape == (3, 2)  # 3 t_values, 2D points


# Test Case 6: Derivative method
def test_derivative_method():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points, curve_type="Bezier")
    t_values = [0.0, 0.5, 1.0]

    # Compute the first derivative of the curve
    derivative_points = curve.derivative(t_values, order=1)

    # Check the returned shape (it should be len(t_values) x d)
    assert derivative_points.shape == (3, 2)  # 3 t_values, 2D points


# Test Case 7: Arc length method
def test_arc_length_method():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points, curve_type="Bezier")

    # Compute the arc length between t_start=0 and t_end=1
    arc_length = curve.arc_length(t_start=0, t_end=1)

    # Check if the arc length is a positive value
    assert arc_length > 0


# Test Case 8: Radius of Curvature (ROC) method
def test_roc_method():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points, curve_type="Bezier")
    t_values = [0.0, 0.5, 1.0]

    # Compute the radius of curvature at the specified t_values
    roc_values = curve.roc(t_values)

    # Check if the ROC values are of the expected shape
    assert roc_values.shape == (3,)


# Test Case 9: Torsion method
def test_torsion_method():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points, curve_type="Bezier")
    t_values = [0.0, 0.5, 1.0]

    # Compute the torsion at the specified t_values
    torsion_values = curve.torsion(t_values)

    # Check if torsion is calculated, should be zero for 2D curves
    assert torsion_values.shape == (3,)
    assert np.all(torsion_values == 0)


# Test Case 10: Dimension property
def test_dimension_property():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points, curve_type="Bezier")

    # Check if dimension property returns the correct value
    assert curve.dimension == 2  # 2D curve


# Test Case 11: String representation
def test_str_method():
    control_points = [(0, 0), (1, 2), (3, 4)]
    curve = Curve(control_points, curve_type="Bezier")

    # Check if the string representation contains the correct curve type and control points count
    assert str(curve) == "Curve type: BezierCurve, Control Points: 3"




import numpy as np
import os
import json
import platform

def export_0d_simulation(tree ,steady=True ,outdir=None ,folder="0d_tmp" ,number_cardiac_cycles=1,flow=None,
                         number_time_pts_per_cycle=5 ,density=1.06 ,viscosity=0.04 ,material="olufsen",
                         olufsen={'k1' :0.0 ,'k2' :-22.5267 ,'k3' :1.0e7 ,'material exponent' :2.0
                                  ,'material pressure' :0.0},
                         linear={'material ehr' :1e7 ,'material pressure' :0.0} ,get_0d_solver=False,
                         path_to_0d_solver=None,viscosity_model='constant' ,vivo=True ,distal_pressure=0,
                         capacitance=True, inductance=True, filename="solver_0d.in", geom_filename="geom.csv"):
    """
    This script builds the 0D input file for running 0D simulation.

    Parameters:
    -----------
                steady : bool (default = True)
                       true/false flag for assigning a time-varying inlet
                       flow. If false, a flow waveform is assigned from
                       empirical data taken from coronary arteries.
                       Refer to svcco.sv_interface.waveform for more details
                outdir : str (default = None)
                       string specifying the output directory for the 0D simulation
                       folder to be created
                folder : str (default = "0d_tmp")
                       string specifying the folder name. This folder will store
                       all 0D simulation files that are generated
                number_cardiac_cycles : int (default = 1)
                       number of cardiac cycles to compute the 0D simulation.
                       This can be useful for obtaining a stable solution.
                number_time_pts_per_cycle : int (default = 5)
                       number of time points to evaluate during the 0D
                       simulation. A default of 5 is good for steady flow cases;
                       however, pulsatile flow will likely require a higher number
                       of time points to adequetly capture the inflow waveform.
                density : float (default = 1.06)
                       density of the fluid within the vascular tree to simulate.
                       By deafult, the density of blood is used.
                       Base Units are in centimeter-grame-second (CGS)
                viscosity : float (default = 0.04)
                       viscosity of the fluid within the vascular tree to simulate.
                       By default, the viscosity if blood is used.
                       Base units are in centimeter-grame-second (CGS)
                material : str (default = "olufsen")
                       string specifying the material model for the vascular
                       wall. Options available: ("olufsen" and "linear")
                olufsen : dict
                         material paramters for olufsen material model
                               'k1' : float (default = 0.0)
                               'k2' : float (default = -22.5267)
                               'k3' : float (default = 1.0e7)
                               'material exponent' : int (default = 2)
                               'material pressure' : float (default = 0.0)
                         By default these parameters numerically yield a "rigid"
                         material model.
                linear : dict
                         material parameters for the linear material model
                               'material ehr' : float (default = 1e7)
                               'material pressure' : float (default = 0.0)

                         By default these parameters numerically yield a "rigid"
                         material model.
                get_0d_solver : bool (default = False)
                         true/false flag to search for the 0D solver. If installed,
                         the solver will be imported into the autogenerated
                         simulation code. Depending on how large the disk-space
                         is for the current machine, this search may be prohibitively
                         expensive if no path is known.
                path_to_0d_solver : str (default = None)
                         string specifying the path to the 0D solver
                viscosity_model : str (default = 'constant')
                         string specifying the viscosity model of the fluid.
                         This may be useful for multiphase fluids (like blood)
                         which have different apparent viscosities at different
                         characteristic length scales. By default, constant
                         viscosity is used.

                         Other viscosity option: 'modified viscosity law'
                              This option leverages work done by Secomb et al.
                              specifically for blood in near-capillary length
                              scales.
                vivo : bool (default = True)
                         true/false flag for using in vivo blood viscosity hematocrit
                         value. If false, in vitro model is used. (Note these
                         are only used for the modified viscosity law model)
                distal_pressure : float (default = 0.0)
                         apply a uniform distal pressure along outlets of the
                         vascular tree
    Returns
    -------
                None
    """
    if outdir is None:
        outdir = os.getcwd() +os.sep +folder
    else:
        outdir = outdir +os.sep +folder
    if not os.path.isdir(outdir):
        os.mkdir(outdir)
    if get_0d_solver:
        #if path_to_0d_solver is None:
        #    path_to_0d_solver = locate_0d_solver()
        #else:
        #    path_to_0d_solver = locate_0d_solver(windows_drive=path_to_0d_solver ,linux_drive=path_to_0d_solver)
        path_to_0d_solver = None
    else:
        path_to_0d_solver = None
    input_file = {'description' :{'description of case' :None,
                                 'analytical results' :None},
                  'boundary_conditions' :[],
                  'junctions' :[],
                  'simulation_parameters' :{},
                  'vessels' :[]}
    simulation_parameters = {}
    simulation_parameters["number_of_cardiac_cycles"] = number_cardiac_cycles
    simulation_parameters["number_of_time_pts_per_cardiac_cycle"] = number_time_pts_per_cycle
    input_file['simulation_parameters'] = simulation_parameters
    terminal_vessels = np.argwhere(tree.data[: ,16 ]==-1).flatten()
    total_outlet_area = np.sum(np.pi *tree.data[terminal_vessels ,21 ]**2)
    total_resistance  = (tree.parameters.terminal_pressure - 1333 * distal_pressure ) /tree.data[0 ,22]
    for vessel in range(tree.data.shape[0]):
        tmp = {}
        tmp['vessel_id'] = vessel
        tmp['vessel_length'] = tree.data[vessel ,20]
        tmp['vessel_name'] = 'branch ' +str(vessel ) +"_seg0"
        tmp['zero_d_element_type'] = "BloodVessel"
        if material == "olufsen":
            material_stiffness = olufsen['k1' ] *np.exp(olufsen['k2' ] *tree.data[vessel ,21] ) +olufsen['k3']
        else:
            material_stiffness = linear['material ehr']
        zero_d_element_values = {}
        if viscosity_model == 'constant':
            nu = tree.parameters.kinematic_viscosity
        elif viscosity_model == 'modified viscosity law':
            W  = 1.1
            lam = 0.5
            D  = tree.data[vessel ,21 ] * 2 *(10000)
            nu_ref = tree.parameters.kinematic_viscosity
            Hd = 0.45  # discharge hematocrit (dimension-less)
            C  = lambda d: (0.8 +np.exp(-0.075) ) *(-1 +( 1 /( 1 +10**(-11 ) * d**12)) ) + \
                        ( 1 /( 1 +10**(-11 ) * d**12))
            nu_vitro_45 = lambda d: 220 *np.exp(-1.3 *d) + 3.2 -2.44 *np.exp(-0.06 * d**0.645)
            nu_vivo_45  = lambda d: 6* np.exp(-0.085 * d) + 3.2 - 2.44 * np.exp(-0.06 * d ** 0.645)
            if vivo == True:
                nu_45 = nu_vivo_45
            else:
                nu_45 = nu_vitro_45
            nu_mod = lambda d: (1 + (nu_45 - 1) * (((1 - Hd) ** C - 1) / ((1 - 0.45) ** C - 1)) * (d / (d - W)) ** (
                        4 * lam)) * (d / (d - W)) ** (4 * (1 - lam))
            ref = nu_ref - nu_mod(10000)  # 1cm (units given in microns)
            nu = nu_mod(D) + ref
        zero_d_element_values["R_poiseuille"] = ((8 * nu / np.pi) * tree.data[vessel, 20]) / tree.data[vessel, 21] ** 4
        if capacitance:
            zero_d_element_values["C"] = (3 * tree.data[vessel, 20] * np.pi * tree.data[vessel, 21] ** 2) / (
                        2 * material_stiffness)
        else:
            zero_d_element_values["C"] = 0.0
        if inductance:
            zero_d_element_values["L"] = (tree.data[vessel, 20] * density) / (np.pi * tree.data[vessel, 21] ** 2)
        else:
            zero_d_element_values["L"] = 0.0
        zero_d_element_values["stenosis_coefficient"] = 0.0
        tmp['zero_d_element_values'] = zero_d_element_values
        if vessel == 0:
            bc = {}
            bc['bc_name'] = "INFLOW"
            bc['bc_type'] = "FLOW"
            bc_values = {}
            if steady:
                if isinstance(flow, type(None)):
                    bc_values["Q"] = [float(tree.data[vessel, 22]), float(tree.data[vessel, 22])]
                    bc_values["t"] = [0, 1]
                else:
                    bc_values["Q"] = [flow, flow]
                    bc_values["t"] = [0, 1]
                with open(outdir + os.sep + "inflow.flow", "w") as file:
                    for i in range(len(bc_values["t"])):
                        file.write("{}  {}\n".format(bc_values["t"][i], bc_values["Q"][i]))
                file.close()
            #else:
            #    time, flow = wave(tree.data[vessel, 22], tree.data[vessel, 21] * 2)  # changed wave function
            #    bc_values["Q"] = flow.tolist()
            #    bc_values["t"] = time.tolist()
            #    bc_values["Q"][-1] = bc_values["Q"][0]
            #    simulation_parameters["number_of_time_pts_per_cardiac_cycle"] = len(bc_values["Q"])
            #    with open(outdir + os.sep + "inflow.flow", "w") as file:
            #        for i in range(len(bc_values["t"])):
            #            file.write("{}  {}\n".format(bc_values["t"][i], bc_values["Q"][i]))
            #    file.close()
            bc['bc_values'] = bc_values
            input_file['boundary_conditions'].append(bc)
            tmp['boundary_conditions'] = {'inlet': "INFLOW"}
            if not np.isnan(tree.data[vessel, 15]) and not np.isnan(tree.data[vessel, 16]):
                junction = {}
                junction['inlet_vessels'] = [vessel]
                junction['junction_name'] = "J" + str(vessel)
                junction['junction_type'] = "NORMAL_JUNCTION"
                junction['outlet_vessels'] = [int(tree.data[vessel, 15]), int(tree.data[vessel, 16])]
                input_file['junctions'].append(junction)
        elif np.isnan(tree.data[vessel, 15]) and np.isnan(tree.data[vessel, 16]):
            bc = {}
            bc['bc_name'] = "OUT" + str(vessel)
            bc['bc_type'] = "RESISTANCE"
            bc_values = {}
            bc_values["Pd"] = 0  # tree.parameters["Pterm"]
            bc_values["R"] = float(total_resistance * (total_outlet_area / (np.pi * tree.data[vessel, 21] ** 2)))
            bc['bc_values'] = bc_values
            input_file['boundary_conditions'].append(bc)
            tmp['boundary_conditions'] = {'outlet': 'OUT' + str(vessel)}
        else:
            junction = {}
            junction['inlet_vessels'] = [vessel]
            junction['junction_name'] = "J" + str(vessel)
            junction['junction_type'] = "NORMAL_JUNCTION"
            junction['outlet_vessels'] = [int(tree.data[vessel, 15]), int(tree.data[vessel, 16])]
            input_file['junctions'].append(junction)
        input_file['vessels'].append(tmp)
    obj = json.dumps(input_file, indent=4)
    with open(outdir + os.sep + filename, "w") as file:
        file.write(obj)
    file.close()

    #with open(outdir + os.sep + "plot_0d_results_to_3d.py", "w") as file:
    #    file.write(make_results)
    #file.close()

    #with open(outdir + os.sep + "plot_0d_results_at_slices.py", "w") as file:
    #    file.write(view_plots)
    #file.close()

    with open(outdir + os.sep + "run.py", "w") as file:
        if platform.system() == "Windows":
            if path_to_0d_solver is not None:
                solver_path = path_to_0d_solver.replace(os.sep, os.sep + os.sep)
            else:
                solver_path = path_to_0d_solver
                print("WARNING: Solver location will have to be given manually")
                print("Current solver path is: {}".format(solver_path))
            solver_file = (outdir + os.sep + filename).replace(os.sep, os.sep + os.sep)
        else:
            if path_to_0d_solver is not None:
                solver_path = path_to_0d_solver
            else:
                solver_path = path_to_0d_solver
                print("WARNING: Solver location will have to be given manually")
                print("Current solver path is: {}".format(solver_path))
            solver_file = outdir + os.sep + "solver_0d.in"
        #file.write(run_0d_script.format(solver_path, solver_file))
    file.close()

    geom = np.zeros((tree.data.shape[0], 8))
    geom[:, 0:3] = tree.data[:, 0:3]
    geom[:, 3:6] = tree.data[:, 3:6]
    geom[:, 6] = tree.data[:, 20]
    geom[:, 7] = tree.data[:, 21]
    np.savetxt(outdir + os.sep + geom_filename, geom, delimiter=",")



from svv.forest.connect.tree_connection import TreeConnection


class ForestConnection:
    def __init__(self, forest, **kwargs):
        self.forest = forest
        self.tree_connections = []
        self.networks = []

    def solve(self, *args, num_vessels=20, attempts=5, **kwargs):
        self.tree_connections = []
        for i in range(self.forest.n_networks):
            tree_connections = TreeConnection(self.forest, i, **kwargs)
            if len(self.tree_connections) > 0:
                other_vessels = []
                for j in range(len(self.tree_connections)):
                    other_vessels.extend(self.tree_connections[j].vessels)
                tree_connections.other_vessels = other_vessels
            tree_connections.solve(*args, num_vessels=num_vessels, attempts=attempts)
            self.tree_connections.append(tree_connections)

    def export_solid(self, cap_resolution=40, extrude_roots=False):
        network_solids = []
        network_lines = []
        network_tubes = []
        for i in range(len(self.tree_connections)):
            network_solid, network_line, network_tube = self.tree_connections[i].export_solid(cap_resolution=cap_resolution,
                                                                                         extrude_roots=extrude_roots)
            network_solids.extend(network_solid)
            network_lines.extend(network_line)
            network_tubes.extend(network_tube)
        return network_solids, network_lines, network_tubes

    def show(self, **kwargs):
        pass